"""
EnsembleBuilder.py
------------------
This module contains reusable builder functions for training ensemble models
as part of the CHURNGUARD churn detection pipeline.

Supported ensemble methods:
1. Voting Classifier (Hard & Soft)
2. Stacking Classifier
3. AdaBoost Classifier
4. Bagging Classifier

Each method returns a dictionary of trained models using preselected base learners.
These models can be evaluated using the ChurnEvaluator class.
"""

# =========================
# IMPORT REQUIRED LIBRARIES
# =========================
from sklearn.ensemble import VotingClassifier, StackingClassifier, AdaBoostClassifier, BaggingClassifier
from sklearn.linear_model import LogisticRegression
from sklearn.tree import DecisionTreeClassifier
from sklearn.naive_bayes import GaussianNB
from sklearn.neighbors import KNeighborsClassifier
from sklearn.svm import SVC
from sklearn.ensemble import RandomForestClassifier
from xgboost import XGBClassifier
from sklearn.pipeline import make_pipeline
from sklearn.preprocessing import StandardScaler


class EnsembleBuilder:
    """
    Class for building ensemble model variants for churn detection.
    """

    def __init__(self, X_train, y_train):
        self.X_train = X_train
        self.y_train = y_train

    def build_voting_classifier(self):
        """
        Build hard and soft voting classifiers using diverse base models.
        Returns:
            dict: trained voting models with names as keys.
        """
        models = {}

        # Define base learners
        lr = LogisticRegression(max_iter=1000)
        rf = RandomForestClassifier(n_estimators=100, random_state=42)
        xgb = XGBClassifier(n_estimators=100, use_label_encoder=False, eval_metric='logloss', random_state=42)

        # Hard Voting
        hard_voting = VotingClassifier(estimators=[
            ('lr', lr), ('rf', rf), ('xgb', xgb)
        ], voting='hard')
        hard_voting.fit(self.X_train, self.y_train)
        models['Voting_Hard'] = hard_voting

        # Soft Voting
        soft_voting = VotingClassifier(estimators=[
            ('lr', lr), ('rf', rf), ('xgb', xgb)
        ], voting='soft')
        soft_voting.fit(self.X_train, self.y_train)
        models['Voting_Soft'] = soft_voting

        return models

    def build_stacking_classifier(self):
        """
        Build a stacking classifier with Logistic Regression as the final estimator.
        Returns:
            dict: trained stacking model with a single variant.
        """
        models = {}

        # Define base learners
        base_learners = [
            ('rf', RandomForestClassifier(n_estimators=100, random_state=42)),
            ('xgb', XGBClassifier(n_estimators=100, use_label_encoder=False, eval_metric='logloss', random_state=42)),
            ('svc', make_pipeline(StandardScaler(), SVC(probability=True)))
        ]

        # Meta-learner
        final_estimator = LogisticRegression()

        stacking = StackingClassifier(estimators=base_learners, final_estimator=final_estimator, passthrough=True)
        stacking.fit(self.X_train, self.y_train)
        models['Stacking'] = stacking

        return models

    def build_adaboost_classifier(self):
        """
        Build AdaBoost classifier variants using different base learners.
        Returns:
            dict: trained AdaBoost models with multiple variants.
        """
        models = {}

        # Variant 1: Default Decision Tree (depth=1)
        ab1 = AdaBoostClassifier(n_estimators=100, random_state=42)
        ab1.fit(self.X_train, self.y_train)
        models['AdaBoost_Default'] = ab1

        # Variant 2: Shallow Decision Tree
        ab2 = AdaBoostClassifier(
            base_estimator=DecisionTreeClassifier(max_depth=2),
            n_estimators=100, random_state=42)
        ab2.fit(self.X_train, self.y_train)
        models['AdaBoost_DT2'] = ab2

        return models

    def build_bagging_classifier(self):
        """
        Build Bagging classifier variants using Decision Tree and KNN.
        Returns:
            dict: trained Bagging models with multiple variants.
        """
        models = {}

        # Variant 1: Decision Tree base learner
        bag1 = BaggingClassifier(
            base_estimator=DecisionTreeClassifier(),
            n_estimators=50, random_state=42)
        bag1.fit(self.X_train, self.y_train)
        models['Bagging_DT'] = bag1

        # Variant 2: K-Nearest Neighbors base learner
        bag2 = BaggingClassifier(
            base_estimator=KNeighborsClassifier(n_neighbors=5),
            n_estimators=50, random_state=42)
        bag2.fit(self.X_train, self.y_train)
        models['Bagging_KNN'] = bag2

        return models
